/**
 * @file main.cpp
 * @brief Disparity map estimation through cost-volume filtering
 * @author Pauline Tan <pauline.tan@ens-cachan.fr>
 *         Pascal Monasse <monasse@imagine.enpc.fr>
 * 
 * Copyright (c) 2012-2013, Pauline Tan, Pascal Monasse
 * All rights reserved.
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * You should have received a copy of the GNU General Pulic License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "costVolume.h"
#include "occlusion.h"
#include "image.h"
#include "cmdLine.h"
#include "io_png.h"
#include <iostream>
#include <fstream> // Charles (ApplyToneCurve)

/// Names of output image files
//static const char* OUTFILE1="disparity.png";
//static const char* OUTFILE2="disparity_occlusion.png";
//static const char* OUTFILE3="disparity_occlusion_filled.png";
//static const char* OUTFILE4="disparity_occlusion_filled_smoothed.png";

static void usage(const char* name) {
	
	std::cerr <<"Guided Filter of a color image.\n"
	<< "Usage: " << name << " [options] image.png \n\n"
	<< "Options (default values in parentheses)\n"
	<< "    -r radius: radius of the guided filter (20)\n"
	<< "    -e epsilon: smothing parameter (25)"
	<< std::endl;
	
	/*
	ParamGuidedFilter p;
    ParamOcclusion q;
    std::cerr <<"Stereo Disparity through Cost Aggregation with Guided Filter\n"
              << "Usage: " << name << " [options] im1.png im2.png dmin dmax\n\n"
              << "Options (default values in parentheses)\n"
              << "Cost-volume filtering parameters:\n"
              << "    -R radius: radius of the guided filter ("
              <<p.kernel_radius << ")\n"
              << "    -A alpha: value of alpha ("<<p.alpha<<")\n"
              << "    -E epsilon: regularization parameter ("<<p.epsilon <<")\n"
              << "    -C tau1: max for color difference ("
              <<p.color_threshold<<")\n"
              << "    -G tau2: max for gradient difference ("
              <<p.gradient_threshold << ")\n\n"
              << "Occlusion detection:\n"
              << "    -o tolDiffDisp: tolerance for left-right disp. diff. ("
              <<q.tol_disp << ")\n\n"
              << "Densification:\n"
              << "    -O sense: fill occlusion, sense='r':right, 'l':left\n"
              << "    -r radius: radius of the weighted median filter ("
              <<q.median_radius << ")\n"
              << "    -c sigmac: value of sigma_color ("
              <<q.sigma_color << ")\n"
              << "    -s sigmas: value of sigma_space ("
              <<q.sigma_space << ")\n\n"
              << "    -a grayMin: value of gray for min disparity (255)\n"
              << "    -b grayMax: value of gray for max disparity (0)"
              << std::endl;
	 */
}

int main(int argc, char *argv[])
{
	int radius=20;
	float epsilon=25.f;
	// char sense='r'; // Camera motion direction: 'r'=to-right, 'l'=to-left
    CmdLine cmd;

	/*
    ParamGuidedFilter paramGF; // Parameters for cost-volume filtering
    cmd.add( make_option('R',paramGF.kernel_radius) );
    cmd.add( make_option('A',paramGF.alpha) );
    cmd.add( make_option('E',paramGF.epsilon) );
    cmd.add( make_option('C',paramGF.color_threshold) );
    cmd.add( make_option('G',paramGF.gradient_threshold) );

    ParamOcclusion paramOcc; // Parameters for filling occlusions
    cmd.add( make_option('o',paramOcc.tol_disp) ); // Detect occlusion
    cmd.add( make_option('O',sense) ); // Fill occlusion
    cmd.add( make_option('r',paramOcc.median_radius) );
    cmd.add( make_option('c',paramOcc.sigma_color) );
    cmd.add( make_option('s',paramOcc.sigma_space) );
	*/
	
    cmd.add( make_option('r',radius) );
	cmd.add( make_option('e',epsilon) );
    try {
        cmd.process(argc, argv);
    } catch(std::string str) {
        std::cerr << "Error: " << str << std::endl<<std::endl;
        usage(argv[0]);
        return 1;
    }
    if(argc!=2) {
        usage(argv[0]);
        return 1;
    }
	// bool detectOcc = cmd.used('o') || cmd.used('O');
	// bool fillOcc = cmd.used('O');
	/*
    if(sense != 'r' && sense != 'l') {
        std::cerr << "Error: invalid camera motion direction " << sense
                  << " (must be r or l)" << std::endl;
        return 1;
	}
	*/
	
	size_t width, height;
	float* pix = io_png_read_f32_rgb(argv[1], &width, &height);
	if (!pix) {
		std::cerr << "Cannot read image file " << argv[1] << std::endl;
		return 1;
	}
	Image im_color(pix, width, height);
	// im_color is 3 channels even if input file is a gray image. (in that case R=G=B).
	
	/*
    // Load images
    size_t width, height, width2, height2;
    float* pix1 = io_png_read_f32_rgb(argv[1], &width, &height);
    float* pix2 = io_png_read_f32_rgb(argv[2], &width2, &height2);
    if(!pix1 || !pix2) {
        std::cerr << "Cannot read image file " << argv[pix1?2:1] << std::endl;
        return 1;
    }
    if(width != width2 || height != height2) {
        std::cerr << "The images must have the same size!" << std::endl;
        return 1;
    }
    Image im1(pix1, width, height);
    Image im2(pix2, width, height);
	*/
	
//	Image boxfilt(width,height);
//	Image tmpfilt(width,height);

	printf("Gray Conversion...\n");
	Image im_gray	 = convert_to_gray(im_color);
	
	printf("Guided Filter on gray image guided by color image.\n");
	Image guidedfilt = guided_filter_color(im_color, im_gray, epsilon*epsilon, radius);
	
	printf("Read LUT file\n");
	float lut[32768];
	std::ifstream lutfile("input_2", std::ios::in);
	if (!lutfile) {
		std::printf("Can't open file: input_2\n");
	}
	else {
		for (int line=0; line<32768; line++) {
			// int max possible value is 32767
			lutfile >> lut[line];
		}
		lutfile.close();
	}
	
	printf("Compute detail and lighting...\n");
	
	Image im_detail(width,height);
	float* pix_lighting = new float[width*height*3];
	Image im_lighting(pix_lighting,width,height);
	float* pix_lighting_ucontrast = new float[width*height*3];
	Image im_lighting_ucontrast(pix_lighting_ucontrast,width,height);
	
	
	Image im_lighting_R = im_lighting.r();
	Image im_lighting_G = im_lighting.g();
	Image im_lighting_B = im_lighting.b();
	Image im_lighting_ucontrast_R = im_lighting_ucontrast.r();
	Image im_lighting_ucontrast_G = im_lighting_ucontrast.g();
	Image im_lighting_ucontrast_B = im_lighting_ucontrast.b();
	Image im_color_R = im_color.r();
	Image im_color_G = im_color.g();
	Image im_color_B = im_color.b();
	
	
	for (size_t y=0; y<height; y++) {
		for (size_t x=0; x<width; x++) {
			im_detail(x,y) = 127.5 + im_gray(x,y) - guidedfilt(x,y);
			
			float gain_det = 0.f;
			float gain_ucont = 0.f;
			if (im_gray(x,y) != 0 ) {
				gain_det   = (im_gray(x,y) +  .6*(im_gray(x,y) - guidedfilt(x,y))) / im_gray(x,y);
				gain_ucont = (im_gray(x,y) + 3.6*(im_gray(x,y) - guidedfilt(x,y))) / im_gray(x,y);
			}
			float gain_lut = 0.f;
			if (im_gray(x,y) != 0 ) {
				int val = int( 128.49803922*im_gray(x,y) );
				gain_lut = lut[int(std::max(0,val))] / 128.49803922 / im_gray(x,y);
			}
			
			im_lighting_R(x,y) = im_color_R(x,y) * gain_lut * gain_det;
			im_lighting_G(x,y) = im_color_G(x,y) * gain_lut * gain_det;
			im_lighting_B(x,y) = im_color_B(x,y) * gain_lut * gain_det;
			
			im_lighting_ucontrast_R(x,y) = im_color_R(x,y) * gain_lut * gain_ucont;
			im_lighting_ucontrast_G(x,y) = im_color_G(x,y) * gain_lut * gain_ucont;
			im_lighting_ucontrast_B(x,y) = im_color_B(x,y) * gain_lut * gain_ucont;
		}
	}
	
	
//	tmpfilt = im.boxFilter(radius);
//	tmpfilt = tmpfilt.boxFilter(radius);
//	boxfilt = tmpfilt.boxFilter(radius);
	
	io_png_write_f32("guidedfilter_base.png", &guidedfilt(0,0), width, height, 1);
	// io_png_write_f32("guidedfilter_gray.png", &im_gray(0,0), width, height, 1);
	io_png_write_f32("guidedfilter_detail.png", &im_detail(0,0), width, height, 1);
	io_png_write_f32("guidedfilter_lighting.png", &im_lighting(0,0), width, height, 3);
	io_png_write_f32("guidedfilter_lighting_ucontrast.png", &im_lighting_ucontrast(0,0), width, height, 3);
	
	
	/*
    // Set disparity range
    int dMin, dMax;
    if(! ((std::istringstream(argv[3])>>dMin).eof() &&
          (std::istringstream(argv[4])>>dMax).eof())) {
        std::cerr << "Error reading dMin or dMax" << std::endl;
        return 1;
    }
    if(dMin>dMax) {
        std::cerr << "Wrong disparity range! (dMin > dMax)" << std::endl;
        return 1;
    }

    Image disp = filter_cost_volume(im1, im2, dMin, dMax, paramGF);
    if(! save_disparity(OUTFILE1, disp, dMin,dMax, grayMin,grayMax)) {
        std::cerr << "Error writing file " << OUTFILE1 << std::endl;
        return 1;
    }

    if(detectOcc) {
        std::cout << "Detect occlusions...";
        Image disp2= filter_cost_volume(im2,im1,-dMax,-dMin,paramGF);
        detect_occlusion(disp, disp2, static_cast<float>(dMin-1),
                         paramOcc.tol_disp);
        if(! save_disparity(OUTFILE2, disp, dMin,dMax, grayMin,grayMax))  {
            std::cerr << "Error writing file " << OUTFILE2 << std::endl;
            return 1;
        }
    }

    if(fillOcc) {
        std::cout << "Post-processing: fill occlusions" << std::endl;
        Image dispDense = disp.clone();
        if(sense == 'r')
            dispDense.fillMaxX(static_cast<float>(dMin));
        else
            dispDense.fillMinX(static_cast<float>(dMin));
        if(! save_disparity(OUTFILE3, dispDense, dMin,dMax, grayMin,grayMax)) {
            std::cerr << "Error writing file " << OUTFILE3 << std::endl;
            return 1;
        }

        std::cout << "Post-processing: smooth the disparity map" << std::endl;
        fill_occlusion(dispDense, im1.medianColor(1),
                       disp, dMin, dMax, paramOcc);
        if(! save_disparity(OUTFILE4, disp, dMin,dMax, grayMin,grayMax)) {
            std::cerr << "Error writing file " << OUTFILE4 << std::endl;
            return 1;
        }
    }
	*/

    free(pix);
	delete [] pix_lighting;
	/*
	free(pix1);
    free(pix2);
	*/
    return 0;
}
