function [img_out,BF,K] = FBFR_FFT(img_in,sigS,sigR,N)
% FBFR_FFT: fast bilateral filter with regression, FFT implementation.
% 
% Usage: img_out = FBFR_FFT(img_in,sigS,sigR,N)
%
% It's an adaptation of the Fast Bilateral Filter of Durand and Dorsey to
% the regression.
%
% The 6 kernels are designed in Fourier domain.
% The convolution is done in Fourier.
% 
% Charles H. - 2016/01/21



% Measurements:
minI = min(img_in(:));
maxI = max(img_in(:));
% Segments
Lg_Seg = (maxI - minI)./(N-1);
Vl_Seg = minI + (0:N-1)*Lg_Seg;

% 9 Kernel in Fourier Domain
[H,W]       = size(img_in);
[fcol,frnk] = meshgrid( [0:(W-1),(-W):(-1)]/(2*W),...
                        [0:(H-1),(-H):(-1)]/(2*H) );
beta    = 2*(pi^2)*(sigS^2);
ker_w   = exp(-beta.*(fcol.^2 + frnk.^2));
ker_w   = ker_w/ker_w(1,1); % normalization
ker_wc  = -1i/2/pi*2*beta.*fcol.*ker_w;
ker_wr  = -1i/2/pi*2*beta.*frnk.*ker_w;
ker_wcc =  (beta/2/pi^2).*(1 - 2*beta*fcol.^2).*ker_w;
ker_wrr =  (beta/2/pi^2).*(1 - 2*beta*frnk.^2).*ker_w;
ker_wcr = -(beta^2/pi^2).*fcol.*frnk.*ker_w;

% Memory
img_wcR  = zeros(H,W);
img_wrR  = zeros(H,W);
img_wccR = zeros(H,W);
img_wcrR = zeros(H,W);
img_wrrR = zeros(H,W);
img_wcF  = zeros(H,W);
img_wrF  = zeros(H,W);
img_wR   = zeros(H,W);
img_wF   = zeros(H,W);

for k=1:N
    
    img_d    = (img_in - Vl_Seg(k));                % d(x): difference
    img_p    = (1 - min(Lg_Seg,abs(img_d))/Lg_Seg); % for interpolation

    img_r    = exp(-(img_d.^2)/(2*sigR^2));         % for normalization
    img_f    = img_in.*img_r;                       % a slice
    
    % Mirror Copy
    img_r_mirror  = cat(1,img_r       ,img_r(H:-1:1,1:+1:W));
    img_r_mirror  = cat(2,img_r_mirror,img_r_mirror(1:+1:end,W:-1:1));
    img_f_mirror  = cat(1,img_f       ,img_f(H:-1:1,1:+1:W));
    img_f_mirror  = cat(2,img_f_mirror,img_f_mirror(1:+1:end,W:-1:1));
    
    img_r_FT = fft2(img_r_mirror);
    img_u_FT = fft2(img_f_mirror);
    
    % Filtering and interpolation between slices
    tmp = real(ifft2(img_r_FT.*ker_w));
    img_wR      = img_wR     + img_p.*tmp(1:H,1:W);
    tmp = real(ifft2(img_u_FT.*ker_w));
    img_wF      = img_wF     + img_p.*tmp(1:H,1:W);
    tmp = real(ifft2(img_r_FT.*ker_wc));
    img_wcR    = img_wcR   + img_p.*tmp(1:H,1:W);
    tmp = real(ifft2(img_r_FT.*ker_wr));
    img_wrR    = img_wrR   + img_p.*tmp(1:H,1:W);
    tmp = real(ifft2(img_r_FT.*ker_wcc));
    img_wccR  = img_wccR + img_p.*tmp(1:H,1:W);
    tmp = real(ifft2(img_r_FT.*ker_wcr));
    img_wcrR  = img_wcrR + img_p.*tmp(1:H,1:W);
    tmp = real(ifft2(img_r_FT.*ker_wrr));
    img_wrrR  = img_wrrR + img_p.*tmp(1:H,1:W);
    tmp = real(ifft2(img_u_FT.*ker_wc));
    img_wcF    = img_wcF   + img_p.*tmp(1:H,1:W);
    tmp = real(ifft2(img_u_FT.*ker_wr));
    img_wrF    = img_wrF   + img_p.*tmp(1:H,1:W);
end

if nargout > 1
    BF = img_wF./img_wR;
    if nargout > 2
        K = img_wR;
    end
end

% Memory
MAT     = zeros(3,3);
VEC     = zeros(3,1);
img_out = zeros(H,W);

% Solve the system at each pixel
for h=1:1:H
    for w=1:1:W
        
        MAT(1,1) = img_wccR(h,w);
        MAT(1,2) = img_wcrR(h,w);
        MAT(1,3) = img_wcR(h,w);
        MAT(2,2) = img_wrrR(h,w);
        MAT(2,3) = img_wrR(h,w);
        MAT(3,3) = img_wR(h,w);

        MAT(2,1) = MAT(1,2);
        MAT(3,1) = MAT(1,3);
        MAT(3,2) = MAT(2,3);

        VEC(1)   = img_wcF(h,w);
        VEC(2)   = img_wrF(h,w);
        VEC(3)   = img_wF(h,w);
        
        % We don't need the slope parameters
        img_out(h,w) = det([MAT(:,1:2) VEC])/det(MAT);
    end
end
end