function diffused = MySelectiveDiffusion(source, filtered, NbIt, sigma)
% Selective Diffusion from Kass & Solomon (Smoothed Local Histogram Filters
% - 2010) with a modification from me.
%
% diffused = MySelectiveDiffusion(source, filtered [, NbIt, sigma] )
%
% This methods correct the over-sharpening of any filter (what I call the
% staircasing in the Bilateral Filter for example) BUT is prone to create
% staircasing effet, as a blend between to image is done and sometimes the
% blending parameter suddenly changes from 0 to 1 (or 1 to 0...). That
% creates this undesirable effect.
% Thus, I introduced a smoothing of this blending parameter \alpha. This is
% the only difference with the original algorithm from Kass & Solomon.
% 
% I also worked to find good parameters to effectively suppress the
% bilateral filter's staircasing while not creating new staircasing (my
% smoothing can't deal with a lot of staircasing).
% I worked with the image "dxo_concorde_gray.png", 680x850 pixels, filtered
% with parameters sig_s = 30, sig_r = .05 (over 1) and NbSlices = 12
% (Durand-Dorsey acceleration with FFTs).
% I found the best parameters are:
%   - nbit  = 6
%   - sigma = 2
%   - lambda = 1
%   - nu = .2
%   - sigma_eps = .8
%
% Where sigma_eps is the standard-deviation for the gaussian I use for
% smoothing the blending parameter \alpha.
%
% The two first parameter can actually be modified, but keep in mind that:
%   - nbit must be inscreased if sigma is set to a smaller value than 2
%   - I don't recommand to use a sigma smaller than 1.5
%   - In return, increasing sigma will allow you to decrease nbit, but I
%     don't recommand values higher than 2.
%
% Charles Hessel -- February 12, 2016

% /// User parameters  ///
if(~exist('NbIt','var')),  NbIt  = 6; end;
if(~exist('sigma','var')), sigma = 2; end;
% /// Fixed parameters ///
lambda    = 1;
nu        = .2;
sigma_eps = .8;
% /// ---------------- ///

Geps = exp(-(-2:2).^2/(2*sigma_eps^2));
Geps = Geps/sum(Geps);

diffused = filtered; clear filtered; % Just rename

for k=1:NbIt
    
    % /// Blur the previous diffused image
    % /// B_{i} = D_{i-1} * G(sig_{i})
    
    r       = max(1,round(3*sigma)); % /// max to avoid r = 0;
    G       = exp(-.5*(-r:r).^2/sigma^2);
    G       = G./sum(G);
    
    B       = conv2(mirror(diffused,[0,r]),G ,'valid');
    B       = conv2(mirror(B       ,[r,0]),G','valid');
    
    % /// Local L^2 deviation computation
    
    sigma_E = sigma*nu;
    r_E     = max(1,round(3*sigma_E)); % /// max to avoid r = 0;
    G_E     = exp(-.5*(-r_E:r_E).^2/(sigma_E)^2);
    G_E     = G_E./sum(G_E);
    
    Eu      = (diffused - source).^2;
    Eu      = conv2(mirror(Eu,[0,r_E]),G_E ,'valid');
    Eu      = conv2(mirror(Eu,[r_E,0]),G_E','valid');
    
    Eb      = (B - source).^2;
    Eb      = conv2(mirror(Eb,[0,r_E]),G_E ,'valid');
    Eb      = conv2(mirror(Eb,[r_E,0]),G_E','valid');
    
    % /// Compute the blending parameter alpha
    
    alpha   = max(0,min(1,  2*(Eb./Eu -.5)  ));
    
    alpha   = conv2(mirror(alpha,[0,2]),Geps ,'valid');
    alpha   = conv2(mirror(alpha,[2,0]),Geps','valid');
    
    % /// Blending: when Eb/Eu > 1, keep D; when Eb/Eu < .5, keep B.
    
    diffused= alpha.*diffused + (1-alpha).*B;
    
    % /// Update sigma
    sigma   = lambda*sigma;
end
end